Files
TheRock/rocm-libraries/projects/rocthrust/thrust/system/cuda/detail/util.h
T
derek.holloway 7502bf64a4
CodeQL / Analyze (actions) (push) Canceled after 0s
CodeQL / Analyze (python) (push) Canceled after 0s
pre-commit / pre-commit (push) Canceled after 0s
Multi-Arch CI / setup (push) Canceled after 0s
Publish build_manylinux_x86_64 images / publish_build_manylinux_x86_64 (push) Canceled after 0s
Setup (Multi-Arch) / setup (push) Canceled after 0s
Multi-Arch CI / Linux::${{ fromJSON(needs.setup.outputs.linux_build_config || '{}').build_variant_label || 'skip' }} (push) Canceled after 0s
Publish TheRock Docker image / build-and-push-image (push) Canceled after 0s
Publish no_rocm_image_ubi10 image / publish_no_rocm_image_ubi10 (push) Canceled after 0s
Multi-Arch CI / Windows::${{ fromJSON(needs.setup.outputs.windows_build_config || '{}').build_variant_label || 'skip' }} (push) Canceled after 0s
Publish TheRock Docker image / build-and-push-image (push) Canceled after 0s
Multi-Arch CI / Manifest Diff (push) Canceled after 0s
TheRock Manifest Diff Report / Generate Manifest Diff Report (push) Canceled after 0s
Multi-Arch CI / CI Summary (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04 images / publish_no_rocm_image_ubuntu24_04 (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04_ocl_rt images / publish_no_rocm_image_ubuntu24_04_ocl_rt (push) Canceled after 0s
Publish TheRock Docker image / build-and-push-image (push) Canceled after 0s
Publish TheRock Docker image / build-and-push-image (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04 images / publish_no_rocm_image_ubuntu24_04_rocgdb (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04_rocgdb images / publish_no_rocm_image_ubuntu24_04_rocgdb (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04 images / publish_no_rocm_image_ubuntu24_04_ocl_rt (push) Canceled after 0s
Publish TheRock Docker image / build-and-push-image (push) Canceled after 0s
Unit Tests / Unit Tests :: ubuntu-24.04 (push) Canceled after 0s
Unit Tests / Unit Tests :: windows-2022 (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04 images / publish_no_rocm_image_ubuntu24_04_media (push) Canceled after 0s
Unit Tests / Unit Tests Summary (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04 images / publish_no_rocm_image_ubuntu24_04_openmpi (push) Canceled after 0s
Flatten subprojects into a single base monorepo
2026-09-02 16:34:47 -07:00

597 lines
16 KiB
C++

/******************************************************************************
* Copyright (c) 2016, NVIDIA CORPORATION. All rights meserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
* * Redistributions of source code must retain the above copyright
* notice, this list of conditions and the following disclaimer.
* * Redistributions in binary form must reproduce the above copyright
* notice, this list of conditions and the following disclaimer in the
* documentation and/or other materials provided with the distribution.
* * Neither the name of the NVIDIA CORPORATION nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL NVIDIA CORPORATION BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************/
#pragma once
#include <thrust/detail/config.h>
#if defined(_CCCL_IMPLICIT_SYSTEM_HEADER_GCC)
# pragma GCC system_header
#elif defined(_CCCL_IMPLICIT_SYSTEM_HEADER_CLANG)
# pragma clang system_header
#elif defined(_CCCL_IMPLICIT_SYSTEM_HEADER_MSVC)
# pragma system_header
#endif // no system header
#include <cub/config.cuh>
#include <cub/detail/device_synchronize.cuh>
#include <cub/util_device.cuh>
#include <thrust/iterator/iterator_traits.h>
#include <thrust/system/cuda/detail/execution_policy.h>
#include <thrust/system/cuda/error.h>
#include <thrust/system_error.h>
#include <cstdio>
#include <exception>
#include <nv/target>
THRUST_NAMESPACE_BEGIN
namespace cuda_cub
{
inline _CCCL_HOST_DEVICE cudaStream_t default_stream()
{
#ifdef CUDA_API_PER_THREAD_DEFAULT_STREAM
return cudaStreamPerThread;
#else
return cudaStreamLegacy;
#endif
}
// Fallback implementation of the customization point.
template <class Derived>
_CCCL_HOST_DEVICE cudaStream_t get_stream(execution_policy<Derived>&)
{
return default_stream();
}
// Entry point/interface.
template <class Derived>
_CCCL_HOST_DEVICE cudaStream_t stream(execution_policy<Derived>& policy)
{
return get_stream(derived_cast(policy));
}
// Fallback implementation of the customization point.
template <class Derived>
_CCCL_HOST_DEVICE bool must_perform_optional_stream_synchronization(execution_policy<Derived>&)
{
return true;
}
// Entry point/interface.
template <class Derived>
_CCCL_HOST_DEVICE bool must_perform_optional_synchronization(execution_policy<Derived>& policy)
{
return must_perform_optional_stream_synchronization(derived_cast(policy));
}
// Fallback implementation of the customization point.
_CCCL_EXEC_CHECK_DISABLE
template <class Derived>
_CCCL_HOST_DEVICE cudaError_t synchronize_stream(execution_policy<Derived>& policy)
{
return cub::SyncStream(stream(policy));
}
// Entry point/interface.
template <class Policy>
_CCCL_HOST_DEVICE cudaError_t synchronize(Policy& policy)
{
return synchronize_stream(derived_cast(policy));
}
// Fallback implementation of the customization point.
_CCCL_EXEC_CHECK_DISABLE
template <class Derived>
_CCCL_HOST_DEVICE cudaError_t synchronize_stream_optional(execution_policy<Derived>& policy)
{
cudaError_t result;
if (must_perform_optional_synchronization(policy))
{
result = synchronize_stream(policy);
}
else
{
result = cudaSuccess;
}
return result;
}
// Entry point/interface.
template <class Policy>
_CCCL_HOST_DEVICE cudaError_t synchronize_optional(Policy& policy)
{
return synchronize_stream_optional(derived_cast(policy));
}
template <class Type>
THRUST_HOST_FUNCTION cudaError_t trivial_copy_from_device(Type* dst, Type const* src, size_t count, cudaStream_t stream)
{
cudaError status = cudaSuccess;
if (count == 0)
{
return status;
}
status = ::cudaMemcpyAsync(dst, src, sizeof(Type) * count, cudaMemcpyDeviceToHost, stream);
cudaStreamSynchronize(stream);
return status;
}
template <class Type>
THRUST_HOST_FUNCTION cudaError_t trivial_copy_to_device(Type* dst, Type const* src, size_t count, cudaStream_t stream)
{
cudaError status = cudaSuccess;
if (count == 0)
{
return status;
}
status = ::cudaMemcpyAsync(dst, src, sizeof(Type) * count, cudaMemcpyHostToDevice, stream);
cudaStreamSynchronize(stream);
return status;
}
template <class Policy, class Type>
THRUST_RUNTIME_FUNCTION cudaError_t
trivial_copy_device_to_device(Policy& policy, Type* dst, Type const* src, size_t count)
{
cudaError_t status = cudaSuccess;
if (count == 0)
{
return status;
}
cudaStream_t stream = cuda_cub::stream(policy);
//
status = ::cudaMemcpyAsync(dst, src, sizeof(Type) * count, cudaMemcpyDeviceToDevice, stream);
cuda_cub::synchronize_optional(policy);
return status;
}
CCCL_DEPRECATED_BECAUSE("Use cuda::std::terminate() instead") inline void _CCCL_HOST_DEVICE terminate()
{
NV_IF_TARGET(NV_IS_HOST, (std::terminate();), (asm("trap;");));
}
_CCCL_HOST_DEVICE inline void throw_on_error(cudaError_t status)
{
// Clear the global CUDA error state which may have been set by the last
// call. Otherwise, errors may "leak" to unrelated kernel launches.
#ifdef THRUST_RDC_ENABLED
cudaGetLastError();
#else
NV_IF_TARGET(NV_IS_HOST, (cudaGetLastError();));
#endif
if (cudaSuccess != status)
{
// Can't use #if inside NV_IF_TARGET, use a temp macro to hoist the device
// instructions out of the target logic.
#ifdef THRUST_RDC_ENABLED
# define THRUST_TEMP_DEVICE_CODE \
printf("Thrust CUDA backend error: %s: %s\n", cudaGetErrorName(status), cudaGetErrorString(status))
#else
# define THRUST_TEMP_DEVICE_CODE printf("Thrust CUDA backend error: %d\n", static_cast<int>(status))
#endif
NV_IF_TARGET(NV_IS_HOST,
(throw thrust::system_error(status, thrust::cuda_category());),
(THRUST_TEMP_DEVICE_CODE; ::cuda::std::terminate();));
#undef THRUST_TEMP_DEVICE_CODE
}
}
_CCCL_HOST_DEVICE inline void throw_on_error(cudaError_t status, char const* msg)
{
// Clear the global CUDA error state which may have been set by the last
// call. Otherwise, errors may "leak" to unrelated kernel launches.
#ifdef THRUST_RDC_ENABLED
cudaGetLastError();
#else
NV_IF_TARGET(NV_IS_HOST, (cudaGetLastError();));
#endif
if (cudaSuccess != status)
{
// Can't use #if inside NV_IF_TARGET, use a temp macro to hoist the device
// instructions out of the target logic.
#ifdef THRUST_RDC_ENABLED
# define THRUST_TEMP_DEVICE_CODE \
printf("Thrust CUDA backend error: %s: %s: %s\n", cudaGetErrorName(status), cudaGetErrorString(status), msg)
#else
# define THRUST_TEMP_DEVICE_CODE printf("Thrust CUDA backend error: %d: %s\n", static_cast<int>(status), msg)
#endif
NV_IF_TARGET(NV_IS_HOST,
(throw thrust::system_error(status, thrust::cuda_category(), msg);),
(THRUST_TEMP_DEVICE_CODE; ::cuda::std::terminate();));
#undef THRUST_TEMP_DEVICE_CODE
}
}
// deprecated [Since 2.8]
template <class ValueType, class InputIt, class UnaryOp>
struct CCCL_DEPRECATED_BECAUSE("Use thrust::transform_iterator") transform_input_iterator_t
{
_CCCL_SUPPRESS_DEPRECATED_PUSH
using self_t = transform_input_iterator_t;
_CCCL_SUPPRESS_DEPRECATED_POP
using difference_type = typename iterator_traits<InputIt>::difference_type;
using value_type = ValueType;
using pointer = void;
using reference = value_type;
using iterator_category = ::cuda::std::random_access_iterator_tag;
InputIt input;
mutable UnaryOp op;
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE transform_input_iterator_t(InputIt input, UnaryOp op)
: input(input)
, op(op)
{}
transform_input_iterator_t(const self_t&) = default;
// UnaryOp might not be copy assignable, such as when it is a lambda. Define
// an explicit copy assignment operator that doesn't try to assign it.
_CCCL_HOST_DEVICE self_t& operator=(const self_t& o)
{
input = o.input;
return *this;
}
/// Postfix increment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator++(int)
{
self_t retval = *this;
++input;
return retval;
}
/// Prefix increment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator++()
{
++input;
return *this;
}
/// Indirection
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE reference operator*() const
{
typename thrust::iterator_value<InputIt>::type x = *input;
return op(x);
}
/// Indirection
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE reference operator*()
{
typename thrust::iterator_value<InputIt>::type x = *input;
return op(x);
}
/// Addition
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator+(difference_type n) const
{
return self_t(input + n, op);
}
/// Addition assignment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t& operator+=(difference_type n)
{
input += n;
return *this;
}
/// Subtraction
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator-(difference_type n) const
{
return self_t(input - n, op);
}
/// Subtraction assignment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t& operator-=(difference_type n)
{
input -= n;
return *this;
}
/// Distance
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE difference_type operator-(self_t other) const
{
return input - other.input;
}
/// Array subscript
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE reference operator[](difference_type n) const
{
return op(input[n]);
}
/// Equal to
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE bool operator==(const self_t& rhs) const
{
return (input == rhs.input);
}
/// Not equal to
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE bool operator!=(const self_t& rhs) const
{
return (input != rhs.input);
}
}; // struct transform_input_iterarot_t
// deprecated [Since 2.8]
template <class ValueType, class InputIt1, class InputIt2, class BinaryOp>
struct CCCL_DEPRECATED_BECAUSE("Use thrust::transform_iterator of a thrust::zip_iterator")
transform_pair_of_input_iterators_t
{
_CCCL_SUPPRESS_DEPRECATED_PUSH
using self_t = transform_pair_of_input_iterators_t;
_CCCL_SUPPRESS_DEPRECATED_POP
using difference_type = typename iterator_traits<InputIt1>::difference_type;
using value_type = ValueType;
using pointer = void;
using reference = value_type;
using iterator_category = ::cuda::std::random_access_iterator_tag;
InputIt1 input1;
InputIt2 input2;
mutable BinaryOp op;
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE
transform_pair_of_input_iterators_t(InputIt1 input1_, InputIt2 input2_, BinaryOp op_)
: input1(input1_)
, input2(input2_)
, op(op_)
{}
transform_pair_of_input_iterators_t(const self_t&) = default;
// BinaryOp might not be copy assignable, such as when it is a lambda.
// Define an explicit copy assignment operator that doesn't try to assign it.
_CCCL_HOST_DEVICE self_t& operator=(const self_t& o)
{
input1 = o.input1;
input2 = o.input2;
return *this;
}
/// Postfix increment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator++(int)
{
self_t retval = *this;
++input1;
++input2;
return retval;
}
/// Prefix increment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator++()
{
++input1;
++input2;
return *this;
}
/// Indirection
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE reference operator*() const
{
return op(*input1, *input2);
}
/// Indirection
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE reference operator*()
{
return op(*input1, *input2);
}
/// Addition
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator+(difference_type n) const
{
return self_t(input1 + n, input2 + n, op);
}
/// Addition assignment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t& operator+=(difference_type n)
{
input1 += n;
input2 += n;
return *this;
}
/// Subtraction
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator-(difference_type n) const
{
return self_t(input1 - n, input2 - n, op);
}
/// Subtraction assignment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t& operator-=(difference_type n)
{
input1 -= n;
input2 -= n;
return *this;
}
/// Distance
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE difference_type operator-(self_t other) const
{
return input1 - other.input1;
}
/// Array subscript
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE reference operator[](difference_type n) const
{
return op(input1[n], input2[n]);
}
/// Equal to
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE bool operator==(const self_t& rhs) const
{
return (input1 == rhs.input1) && (input2 == rhs.input2);
}
/// Not equal to
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE bool operator!=(const self_t& rhs) const
{
return (input1 != rhs.input1) || (input2 != rhs.input2);
}
}; // struct transform_pair_of_input_iterators_t
// deprecated [Since 2.8]
struct CCCL_DEPRECATED_BECAUSE("Use cuda::std::identity") identity
{
template <class T>
_CCCL_HOST_DEVICE T const& operator()(T const& t) const
{
return t;
}
template <class T>
_CCCL_HOST_DEVICE T& operator()(T& t) const
{
return t;
}
};
// deprecated [Since 2.8]
template <class T>
struct CCCL_DEPRECATED_BECAUSE("Use thrust::counting_iterator") counting_iterator_t
{
_CCCL_SUPPRESS_DEPRECATED_PUSH
using self_t = counting_iterator_t;
_CCCL_SUPPRESS_DEPRECATED_POP
using difference_type = T;
using value_type = T;
using pointer = void;
using reference = T;
using iterator_category = ::cuda::std::random_access_iterator_tag;
T count;
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE counting_iterator_t(T count_)
: count(count_)
{}
/// Postfix increment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator++(int)
{
self_t retval = *this;
++count;
return retval;
}
/// Prefix increment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator++()
{
++count;
return *this;
}
/// Indirection
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE reference operator*() const
{
return count;
}
/// Indirection
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE reference operator*()
{
return count;
}
/// Addition
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator+(difference_type n) const
{
return self_t(count + n);
}
/// Addition assignment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t& operator+=(difference_type n)
{
count += n;
return *this;
}
/// Subtraction
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t operator-(difference_type n) const
{
return self_t(count - n);
}
/// Subtraction assignment
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE self_t& operator-=(difference_type n)
{
count -= n;
return *this;
}
/// Distance
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE difference_type operator-(self_t other) const
{
return count - other.count;
}
/// Array subscript
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE reference operator[](difference_type n) const
{
return count + n;
}
/// Equal to
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE bool operator==(const self_t& rhs) const
{
return (count == rhs.count);
}
/// Not equal to
_CCCL_HOST_DEVICE _CCCL_FORCEINLINE bool operator!=(const self_t& rhs) const
{
return (count != rhs.count);
}
}; // struct count_iterator_t
} // namespace cuda_cub
THRUST_NAMESPACE_END